package biz.hammurapi.configx;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.w3c.dom.Document;

import biz.hammurapi.config.Command;
import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.config.MapContext;
import biz.hammurapi.config.RestartCommand;
import biz.hammurapi.config.Restartable;
import biz.hammurapi.config.RuntimeConfigurationException;
import biz.hammurapi.configx.xmltypes.ComponentDocument;
import biz.hammurapi.configx.xmltypes.NamedObjectSpecification;
import biz.hammurapi.configx.xmltypes.ObjectSpecification;
import biz.hammurapi.configx.xmltypes.Path;
import biz.hammurapi.configx.xmltypes.Property;
import biz.hammurapi.configx.xmltypes.Typed;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.xml.dom.DOMUtils;

/**
 * This class instantiates Java objects from XML configuration
 * parsed by XML Beans.
 * 
 * This class is a counterpart and replacement for biz.hammurapi.config.DomConfigFactory
 * @author Pavel
 *
 */
public class XmlConfigFactory {
	
	public static final String XML_EXTENSION = ".xml";
	public static final String CONFIG_RESOURCE_PREFIX = "META-INF/config/";
	
	private static final Logger logger = Logger.getLogger(XmlConfigFactory.class.getName());

	/**
	 * Creates class loader from specification
	 * @param spec Specification. Can be null.
	 * @param parent Parent classloader. Can be null.
	 * @param context Context URL. Can be null.
	 * @return New class loader if spec is not null, or parent if it is not null or classloader used to load XmlConfigFactory class.
	 * @throws MalformedURLException
	 */
	public static ClassLoader getClassLoader(Path spec, ClassLoader parent, URL context) throws MalformedURLException {
		ClassLoader ret = parent==null ? XmlConfigFactory.class.getClassLoader() : parent;
		
		ArrayList<URL> elements = new ArrayList<URL>();
		collectElements(spec, elements, context);
		if (!elements.isEmpty()) {			
			ret = new URLClassLoader(elements.toArray(new URL[elements.size()]), ret);
		}
		
		return ret;
	}
	
	private static void collectElements(Path path, List<URL> receiver, URL context) throws MalformedURLException {
		if (path!=null) {
			if (path.getContextUri()!=null) {
				context = new URL(context, path.getContextUri());
			}
			
			String[] pea = path.getPathElementArray();
			for (int i=0; i<pea.length; ++i) {
				receiver.add(new URL(context, pea[i]));
			}
			
			Path[] pa = path.getPathArray();
			for (int i=0; i<pa.length; ++i) {
				collectElements(pa[i], receiver, context);
			}
		}
	}
	
	/**
	 * Validates specification and then creates object. Use this method on 
	 * top-level elements.
	 * @param spec
	 * @param classLoader
	 * @param context
	 * @return
	 * @throws ConfigurationException
	 */
	public static Object validateAndCreate(Typed spec, ClassLoader classLoader, URL context) throws ConfigurationException {		
		ArrayList<?> validationErrors = new ArrayList<Object>();
		XmlOptions validationOptions = new XmlOptions();
		validationOptions.setErrorListener(validationErrors);
		if (!spec.validate(validationOptions)) {
			StringBuffer eb = new StringBuffer("Invalid specification: ");
		    for (Object err: validationErrors) {
		        eb.append("\n\t>> " + err);
		    }
			
			throw new ConfigurationException(eb.toString());
		}
		
		return create(spec, classLoader, context);
	}	
	
	/**
	 * Instantiates object from Typed XML definition or its subclass.
	 * @param spec Specification
	 * @param classLoader Class loader. If it is null, then factory's classloader is used.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static <T> T create(Typed spec, ClassLoader classLoader, URL context) throws ConfigurationException {		
		
		if (spec.getType()==null) {
			throw new ConfigurationException("Type is not set");
		}		
		
		try {
			
			if (spec instanceof ObjectSpecification) {
				biz.hammurapi.configx.xmltypes.ObjectSpecification.Constructor constructorSpec = ((ObjectSpecification) spec).getConstructor();
				if (constructorSpec!=null) {
					throw new UnsupportedOperationException("Constructors are not yet supported");
				}
			}
			
			ClassLoader theClassLoader = getClassLoader(spec.getClasspath(), classLoader, context);
			
			Class<T> clazz = (Class<T>) theClassLoader.loadClass(spec.getType());
			if (XmlConfigurable.class.isAssignableFrom(clazz)) {
				T ret = clazz.newInstance();
				((XmlConfigurable) ret).configure(spec);
				return ret;
			} 
				
			Constructor<T> candidate = null;
			Constructor<T>[] ca = clazz.getConstructors();
			for (int i = 0; i<ca.length; ++i) {
				Class<?>[] parameterTypes = ca[i].getParameterTypes();
				if (parameterTypes.length==1 
						&& parameterTypes[0].isInstance(spec) 
						&& (candidate==null || candidate.getParameterTypes()[0].isAssignableFrom(parameterTypes[0]))) {
					candidate = ca[i];
				}
			}
			
			if (candidate!=null) {
				return candidate.newInstance(new Object[] {spec});
			}
			
			T ret = clazz.newInstance();
			if (spec instanceof ObjectSpecification) {
				Map<String, Object> properties = instantiate(((ObjectSpecification) spec).getPropertyArray());
				if (!properties.isEmpty()) {
					DomConfigFactory.inject(ret, new MapContext(properties));
				}
				Map<String, Object> attributes = instantiate(((ObjectSpecification) spec).getAttributeArray(), theClassLoader, context);
				if (!attributes.isEmpty()) {
					DomConfigFactory.inject(ret, new MapContext(attributes));
				}
			}
			
			return ret;			
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException("Could not instantiate object from specification: "+e, e);
		}		
	}	

	/**
	 * Instantiates properties from XML defintion
	 * @param property
	 * @return
	 * @throws ConfigurationException
	 */
	public static Hashtable<String, Object> instantiate(Property[] properties) throws ConfigurationException {
		/**
		 * Marker class for multi-value entries.
		 * @author Pavel
		 */
		class MultiValueList extends ArrayList<Object> {
			
		}
		
		Hashtable<String, Object> ret = new Hashtable<String, Object>();
		if (properties!=null) {
			for (int i=0; i<properties.length; ++i) {
				String pName = properties[i].getName();
				Object pValue = properties[i].getStringValue();
				String pType = properties[i].getType();
				if (pType!=null && pType.trim().length()!=0) {
					try {
						pValue = ConvertingService.convert(pValue, Class.forName(pType));
					} catch (ClassNotFoundException e) {
						throw new ConfigurationException("Cannot instantiate property of type "+properties[i].getType()+" from value "+properties[i].getStringValue());
					}
				}
				
				Object eValue = ret.get(pName);
				if (eValue==null) {
					ret.put(pName, pValue);
				} else if (eValue instanceof MultiValueList) {
					((MultiValueList) eValue).add(pValue);
				} else {
					MultiValueList mvl = new MultiValueList();
					mvl.add(eValue);
					mvl.add(pValue);
					ret.put(pName, mvl);
				}
			}
		}
		return ret;
	}	

	/**
	 * Instantiates properties from XML defintion
	 * @param property
	 * @return
	 * @throws ConfigurationException
	 */
	public static Hashtable<String, Object> instantiate(NamedObjectSpecification[] attributes, ClassLoader classLoader, URL context) throws ConfigurationException {
		/**
		 * Marker class for multi-value entries.
		 * @author Pavel
		 */
		class MultiValueList extends ArrayList<Object> {
			
		}
		
		Hashtable<String, Object> ret = new Hashtable<String, Object>();
		if (attributes!=null) {
			for (int i=0; i<attributes.length; ++i) {
				Object aValue = create(attributes[i], classLoader, context);
				String aName = attributes[i].getName();
				
				Object eValue = ret.get(aName);
				if (eValue==null) {
					ret.put(aName, aValue);
				} else if (eValue instanceof MultiValueList) {
					((MultiValueList) eValue).add(aValue);
				} else {
					MultiValueList mvl = new MultiValueList();
					mvl.add(eValue);
					mvl.add(aValue);
					ret.put(aName, mvl);
				}
			}
		}
		return ret;
	}
	
	private static Object create(Document doc) throws XmlException, ConfigurationException {
		XmlObject spec = XmlObject.Factory.parse(doc.getDocumentElement());
		ArrayList<?> validationErrors = new ArrayList<Object>();
		XmlOptions validationOptions = new XmlOptions();
		validationOptions.setErrorListener(validationErrors);
		if (!spec.validate(validationOptions)) {
			logger.severe("Configuration validation failed:");
		    for (Object err: validationErrors) {
		        logger.severe("\t>> " + err);
		    }
			
			System.exit(2);
		}
		if (spec instanceof Typed) {
			return create((Typed) spec, null, null);
		} 
		
		logger.severe("Invalid configuration, expected typed element");
		System.exit(1);
		return null;	
	}

	/**
	 * @param args
	 */
	public static void main(final String[] args) {
		final long start=System.currentTimeMillis();
		if (args.length==0) {
			logger.info("Usage: java <options> "+XmlConfigFactory.class.getName()+" <configuration URL> [<additional parameters>]");
			System.exit(1);
		}
		
		final boolean stopInHook = "yes".equalsIgnoreCase(System.getProperty("biz.hammurapi.configx.XmlFactory:shutdownHook"));
		final Object[] oa = {null};
		
		if (stopInHook) {
			Runtime.getRuntime().addShutdownHook(
					new Thread() {
						public void run() {
							if (oa[0] instanceof Component) {
								try {							
									((Component) oa[0]).stop();
								} catch (ConfigurationException e) {
									logger.severe("Could not properly stop "+oa[0]);
									e.printStackTrace();
								}
								logger.info("Total execution time: "+((System.currentTimeMillis()-start)/1000)+" sec.");
							}
						}
					});			
		}
						
		RestartCommand run = new RestartCommand() {
			
			int attempt;
			
			public void run() {
				try {
					if (attempt > 0) {	
						long restartDelay = getRestartDelay();
						logger.info("Restarting in "+restartDelay+" milliseconds. Attempt " + (attempt+1));
						try {
							Thread.sleep(restartDelay);
						} catch (InterruptedException ie) {
							ie.printStackTrace();
							System.exit(4);
						}
					}
					
					++attempt;
					
					if (args[0].startsWith("url:")) {
						oa[0] = create(DOMUtils.parse(new URL(args[0].substring("url:".length())).openStream()));
					} else if (args[0].startsWith(biz.hammurapi.config.DomConfigFactory.RESOURCE_PREFIX)) {
						InputStream stream = XmlConfigFactory.class.getClassLoader().getResourceAsStream(args[0].substring(biz.hammurapi.config.DomConfigFactory.RESOURCE_PREFIX.length()));
						if (stream==null) {
							logger.severe("Resource does not exist.");
							System.exit(1);
						}
						oa[0] = create(DOMUtils.parse(stream));
					} else {
						File file = new File(args[0]);
						if (!file.exists()) {
							logger.severe("File does not exist or not a file.");
							System.exit(1);
						}
						if (!file.isFile()) {
							logger.severe("Not a file.");
							System.exit(1);
						}
						oa[0] = create(DOMUtils.parse(file));
					}

					if (oa[0] instanceof Component) {
						((Component) oa[0]).start();
					}
					
					if (oa[0] instanceof Restartable) {
						((Restartable) oa[0]).setRestartCommand(this);
					}
					
					try {
						if (oa[0] instanceof Context) {
							Context container=(Context) oa[0];
							for (int i=1; i<args.length; i++) {
								Object toExecute=container.get(args[i]);
								if (toExecute instanceof Command) {
									((Command) toExecute).execute(args);
								} else if (toExecute==null) {						
									logger.warning("Name not found: " +args[i]);
								} else {
									logger.warning("Not executable: (" +args[i]+") "+toExecute.getClass().getName());						
								}
							}					
						} else if (oa[0] instanceof Command) {
							((Command) oa[0]).execute(args);
						}
					} finally {			
						if (oa[0] instanceof Component) {
							if (!stopInHook) {
								((Component) oa[0]).stop();
							}
						}
					}
				} catch (MalformedURLException e) {
					logger.severe("Bad configuration URL: "+args[0]);
					System.exit(2);
				} catch (Exception e) {
					e.printStackTrace();
					if (attempt > 1) {	
						if (oa[0] instanceof Component) {
							try {
								((Component) oa[0]).stop();
							} catch (Exception ex) {
								logger.severe("Cannot stop component before restart: "+e);
								ex.printStackTrace();
							}
						}
						new Thread(this, "Restart thread "+getAttempt()).start(); // Use a new thread to avoid stack overflowing in the case of too many attempts.
					} else {					
						System.exit(3);
					}
				}
			}

			public int getAttempt() {
				return attempt;
			}

			public long getRestartDelay() {
				String rd = System.getProperty(RESTART_DELAY_PROPERTY);
				if (rd!=null) {
					try {
						return Long.parseLong(rd);
					} catch (NumberFormatException e) {
						// Ignore
					}
				}
				
				return DEFAULT_RESTART_DELAY;
			}
		};
		
		run.run();
		
	}

	/**
	 * Parses component document (as defined in configx.xsd) at URL and creates object from it.
	 * @param <T>
	 * @param url
	 * @param classLoader
	 * @return
	 * @throws ConfigurationException 
	 */
	public static Object create(URL configURL, ClassLoader classLoader) throws ConfigurationException {
		try {
			ComponentDocument cd = ComponentDocument.Factory.parse(configURL);
			return create(cd, classLoader, configURL);
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
	}

	/**
	 * Parses component document (as defined in configx.xsd) at URL and creates object from it.
	 * @param <T>
	 * @param Configuration document reader.
	 * @param classLoader
	 * @return Configured object.
	 * @throws ConfigurationException 
	 */
	public static Object create(Reader reader, ClassLoader classLoader) throws ConfigurationException {
		try {
			ComponentDocument cd = ComponentDocument.Factory.parse(reader);
			return create(cd, classLoader, null);
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
	}

	public static Object create(
			ComponentDocument cd, 
			ClassLoader classLoader,
			URL baseURL) 
	throws ConfigurationException {
		
		ArrayList<?> validationErrors = new ArrayList<Object>();
		XmlOptions validationOptions = new XmlOptions();
		validationOptions.setErrorListener(validationErrors);
		if (!cd.validate(validationOptions)) {
			StringBuffer eb = new StringBuffer("Invalid definition: ");
		    for (Object err: validationErrors) {
		        eb.append("\n\t>> " + err);
		    }
			
			throw new ConfigurationException(eb.toString());
		}
		
		return create(cd.getComponent(), classLoader, baseURL);
	}
	
	/**
	 * Loads providers for a given service. Providers are loaded anew at every
	 * invocation of this method.
	 * @param <T>
	 * @param service Service class.
	 * @param classLoader Class loader
	 * @return Iterator of providers of given class.
	 * @throws ConfigurationException
	 */
	public static <T> Iterator<T> loadProviders(final Class<T> service, final ClassLoader classLoader) {
		final String resName = CONFIG_RESOURCE_PREFIX+service.getName()+XML_EXTENSION;
		return new Iterator<T>() {
			
			private Enumeration<URL> resources;
			
			private synchronized Enumeration<URL> getResources() {
				if (resources==null) {
					try {
						resources = classLoader==null ? ClassLoader.getSystemResources(resName) : classLoader.getResources(resName);						
					} catch (IOException e) {
						throw new RuntimeConfigurationException(e);
					}					
				}
				return resources;
			}

			public boolean hasNext() {
				return getResources().hasMoreElements();
			}

			@SuppressWarnings("unchecked")
			public T next() {
				try {
					return (T) create(getResources().nextElement(), classLoader);
				} catch (ConfigurationException e) {
					throw new RuntimeConfigurationException(e);
				}
			}

			public void remove() {
				throw new UnsupportedOperationException("This operation is not supported");				
			}
			
		};
		
	}
	
	private static class ServiceEntry<T> {
		Class<T> clazz;
		ClassLoader classLoader;
		
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result	+ ((classLoader == null) ? 0 : classLoader.hashCode());
			result = prime * result + ((clazz == null) ? 0 : clazz.hashCode());
			return result;
		}
		
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			
			ServiceEntry<T> other = (ServiceEntry<T>) obj;
			if (classLoader == null) {
				if (other.classLoader != null)
					return false;
			} else if (!classLoader.equals(other.classLoader))
				return false;
			if (clazz == null) {
				if (other.clazz != null)
					return false;
			} else if (!clazz.equals(other.clazz))
				return false;
			return true;
		}		
	}
	
	private static final Map<ServiceEntry<?>, Iterable<?>> providers = new HashMap<ServiceEntry<?>, Iterable<?>>();
	
	/**
	 * Loads and caches providers. If provider implements Component, then its start
	 * method is invoked after creation, and stop() method is invoked in shutdown hook.
	 * @param <T>
	 * @param service
	 * @param classLoader
	 * @return
	 * @throws ConfigurationException 
	 */
	@SuppressWarnings("unchecked")
	public static <T> Iterable<T> providers(final Class<T> service, final ClassLoader classLoader) throws ConfigurationException {
		synchronized (providers) {
			ServiceEntry<T> key = new ServiceEntry<T>();
			key.clazz = service;
			key.classLoader = classLoader;
			
			Iterable<T> ret = (Iterable<T>) providers.get(key);
			if (ret==null) {
				ArrayList<T> rt = new ArrayList<T>();
				Iterator<T> pIt = loadProviders(service, classLoader);
				while (pIt.hasNext()) {
					T nextProvider = pIt.next();
					if (nextProvider instanceof Component) {
						((Component) nextProvider).start();
					}
					rt.add(nextProvider);
				}
				
				if (!rt.isEmpty()) {
					if (providers.isEmpty()) {
						Runtime.getRuntime().addShutdownHook(new Thread() {
							
							@Override
							public void run() {
								synchronized (providers) {
									for (Iterable<?> typeProviders: providers.values()) {
										for (Object typeProvider: typeProviders) {
											if (typeProvider instanceof Component) {
												try {
													((Component) typeProvider).stop();
												} catch (Exception e) {
													logger.log(Level.WARNING, "Could not stop service provider "+typeProvider+": "+e, e);
												}
											}
										}
									}
								}
							}
						});
					}
				}
				
				ret=rt;
				providers.put(key, ret);
			}			
			return ret;
		}

	}
	
	private static Map<ServiceEntry<?>, Object> services = new HashMap<ServiceEntry<?>, Object>();

	/**
	 * Returns composite service. In order to use this method providers shall implement
	 * CompositeProvider interface.
	 * @param <T>
	 * @param service
	 * @return
	 * @throws ConfigurationException 
	 */
	@SuppressWarnings("unchecked")
	public static <T> T getService(Class<T> service, ClassLoader classLoader) throws ConfigurationException {
		synchronized (services) {
			ServiceEntry<T> key = new ServiceEntry<T>();
			key.clazz = service;
			key.classLoader = classLoader;
			T ret = (T) services.get(key);
			if (ret == null) {
				for (T provider: providers(service, classLoader)) {
					if (ret==null) {
						ret = provider;
					} else if (ret instanceof CompositeProvider) {
						((CompositeProvider) ret).merge(provider);
					} else if (provider instanceof CompositeProvider) {
						((CompositeProvider) provider).merge(ret);
						ret = provider;
					} else {
						throw new ConfigurationException("Cannot merge providers for service "+service+" - neither of them implements " + CompositeProvider.class.getName());
					}
				}
				services.put(key, ret);
			}
			return ret;
		}
	}
}
