package biz.hammurapi.metrics.jmx;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.StandardMBean;

import biz.hammurapi.metrics.MeasurementCategoryFactory;
import biz.hammurapi.metrics.MeasurementConsumer;
import biz.hammurapi.metrics.SimpleSlice;
import biz.hammurapi.metrics.Slice;

public abstract class JmxMeasurementCategoryFactory extends MeasurementCategoryFactory {
	private String category;
	private Integer interval;
	
	public void setCategory(String category) {
		this.category = category;
	}
	
	public void setInterval(int pInterval) {
		this.interval = pInterval;
	}
	
	private Map<String, MeasurementConsumer> measurementConsumers = new HashMap<String, MeasurementConsumer>();
	
	private class JmxSliceImpl implements JmxSlice {
		
		private String name;
		private Slice data;
		private long to;
		private Slice accumulator;

		public JmxSliceImpl(String name) {
			this.name = name;
			data = new SimpleSlice(name);
			if (interval!=null) {
				accumulator = new SimpleSlice(name);
				to = System.currentTimeMillis();
			}
		}

		public double getAverage() {
			return getData().getAvg();
		}

		public double getDeviation() {
			return getData().getDeviation();
		}

		public Date getFrom() {
			return new Date(getData().getFrom());
		}

		public double getMax() {
			return getData().getMax();
		}

		public double getMin() {
			return getData().getMin();
		}

		public int getNumber() {
			return getData().getNumber();
		}

		public Date getTo() {
			return new Date(getData().getTo());
		}

		public double getTotal() {
			return getData().getTotal();
		}

		public void reset() {
			data = new SimpleSlice(name);
		}
		
		Slice getData() {
			checkTime();
			return data;
		}

		private void checkTime() {
			if (interval!=null) {
				long now = System.currentTimeMillis();
				if (now > to) {
					data = accumulator;
					accumulator = new SimpleSlice(name);
					while (to<now) {
						to += interval;
					}
				}
			}
		}
		
		Slice getAccumulator() {
			checkTime();
			return interval==null ? data : accumulator;
		}
		
	}

	public synchronized MeasurementConsumer getMeasurementConsumer(final String categoryName) {
		MBeanServer mBeanServer = getMBeanServer();
		if (mBeanServer!=null 
				&& (category==null || category.equals(categoryName) || (categoryName!=null && categoryName.startsWith(category+".")))) {
			MeasurementConsumer ret = (MeasurementConsumer) measurementConsumers.get(categoryName);
			if (ret==null) {								
				ret = new MeasurementConsumer() {
					private Map<String, JmxSliceImpl> sliceMap = new HashMap<String, JmxSliceImpl>();

					public synchronized void addMeasurement(String name, double value, long time) {
						JmxSliceImpl slice = sliceMap.get(name);
						if (slice==null) {
							slice=new JmxSliceImpl(name);
							sliceMap.put(name, slice);
							
							try {
					            MBeanServer mbs = getMBeanServer();
					            if (mbs!=null) {
						            mbs.registerMBean(new StandardMBean(slice, JmxSlice.class), objectName(categoryName, name));
					            }
							} catch (Exception e) {
								System.err.println("[JmxMeasurementCategoryFactory] ERROR: "+e);
								return;
							}
						}
						slice.getAccumulator().add(value, time);
					}
					
				};				
			}
			return ret;		
		}
		return null;
	}
	
	/**
	 * This implementation tries to use Java 5 management features.
	 * Returns null if JVM is not of version 5.
	 * @return MBean server to register MBeans with
	 */
	protected abstract MBeanServer getMBeanServer();

	/**
	 * Constructs MBean name from category name and measurement name. 
	 * @param categoryName
	 * @param name
	 * @return
	 * @throws NullPointerException 
	 * @throws MalformedObjectNameException 
	 */
	protected ObjectName objectName(final String categoryName, String name) throws MalformedObjectNameException {
		return new ObjectName("biz.hammurapi.metrics:category="+categoryName+",name="+name);
	}

}
