package biz.hammurapi.registry;

import java.util.List;
import java.util.Set;

/**
 * Entry in Java Service/Component Registry.
 * @author Tatyana Konukova
 *
 */
public interface RegistryEntry {
	
	/**
	 * @return Component/service name
	 */
	String getName();
	
	/**
	 * @return Short description to display in component selection tree control.
	 */
	String getTooltip();

	/**
	 * @return Long description to display in component details panel.
	 */
	String getDescription();
	
	/**
	 * One entry may produce different types of objects. This method returns
	 * types supported by the entry. 
	 * @return Classes/interfaces implemented by this entry.
	 */
	Class<?>[] getComponentTypes();
	
	/**
	 * One entry may support different factory types, e.g. CodeSnippet factory 
	 * creates a code snippet. That code snippet instantiates and configures component.
	 * XML document factory creates an XML element/document for instantiating/configuring
	 * component.
	 * @return
	 */
	Class<?>[] getFactoryTypes();	

	/**
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of component types are displayed.
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of factory types are displayed.
	 * @param filterTokens Filter tokens
	 * @return True if entry itself matches (i.e. it can be selected in the browser). 
	 */
	boolean match(Class<?>[] componentTypes, Class<?> factoryTypes, Set<String> filterTokens);
	
	/**
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of component types are displayed.
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of factory types are displayed.
	 * @param filterTokens Filter tokens
	 * @return True if entry has children matching tokens.
	 */
	boolean hasChildren(Class<?>[] componentTypes, Class<?> factoryTypes, Set<String> filterTokens);
	
	/**
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of component types are returned.
	 * @param componentTypes If this parameter is not null then only entries which
	 * support one of factory types are returned.
	 * @param filterTokens
	 * @return Children matching tokens.
	 */
	List<RegistryEntry> getChildren(Class<?>[] componentTypes, Class<?> factoryTypes, Set<String> filterTokens);

	 /**
	  * Creates new factory for configuring and instantiating registry entry of given type.
	  * @param <T> Factory class.
	  * @return Configurator for the factory.
	  */
	 <T> FactoryConfigurator<T> newFactory(Class<T> factoryClass);
}
