package biz.hammurapi.remoting.http;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpConnectionManager;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;

import biz.hammurapi.remoting.Invocation;

public class HttpRemoteInvocationHandler implements InvocationHandler {
	
	private HttpReference reference;
	private HttpConnectionManager connectionManager;
	private String cookie;
	private ClassLoader classLoader;

	public HttpRemoteInvocationHandler(HttpReference reference, 
			String cookie, 
			HttpConnectionManager connectionManager,
			ClassLoader classLoader) {
		this.reference = reference;
		this.connectionManager = connectionManager;
		this.cookie = cookie;
		this.classLoader = classLoader;
	}
	
	/**
	 * Used for remote class and resource loading.
	 * @param name
	 * @return
	 */
	public byte[] getResource(String resourceName) {
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(new GZIPOutputStream(baos));
			oos.writeObject(resourceName);
			oos.close();
			HttpClient httpClient = new HttpClient(connectionManager);						
			PostMethod httpMethod = new PostMethod(reference.getRemoteUrl());
			try {
				httpMethod.setRequestEntity(new ByteArrayRequestEntity(baos.toByteArray()));
					if (cookie!=null) {
						httpMethod.setRequestHeader("Cookie", cookie);
					}
					int statusCode = httpClient.executeMethod(httpMethod);
					if (statusCode == HttpStatus.SC_OK) {
						return httpMethod.getResponseBody();
					} else if (statusCode == HttpStatus.SC_NOT_FOUND) {
						return null;
					}
					throw new HttpRemoteException("Remoting problem, HTTP error code: "+statusCode+", "+httpMethod.getResponseBodyAsString());
			} finally {
				httpMethod.releaseConnection();
			}										
		} catch (HttpRemoteException e) {
			throw e;					
		} catch (Exception e) {
			throw new HttpRemoteException("Remoting problem: "+e, e);
		}
	}

	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
		try {
			Invocation invocation = new Invocation(null, method, args, reference.getId());
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(new GZIPOutputStream(baos)) {
				
				{
					enableReplaceObject(true);
				}
				
				protected Object replaceObject(Object obj) throws IOException {
					if (Proxy.isProxyClass(obj.getClass())) {
						InvocationHandler ih = Proxy.getInvocationHandler(obj);
						if (ih instanceof HttpRemoteInvocationHandler) {
							return ((HttpRemoteInvocationHandler) ih).reference;
						}
					}
					return super.replaceObject(obj);
				}
			};
			oos.writeObject(invocation);
			oos.close();
			
			HttpClient httpClient = new HttpClient(connectionManager);						
			PostMethod httpMethod = new PostMethod(reference.getRemoteUrl());
			try {
				httpMethod.setRequestEntity(new ByteArrayRequestEntity(baos.toByteArray()));
					if (cookie!=null) {
						httpMethod.setRequestHeader("Cookie", cookie);
					}
					int statusCode = httpClient.executeMethod(httpMethod);
					if (statusCode == HttpStatus.SC_OK) {
						ObjectInputStream ois = new ObjectInputStream(new GZIPInputStream(new ByteArrayInputStream(httpMethod.getResponseBody()))) {
							
							{
								enableResolveObject(true);
							}
							
							protected Object resolveObject(Object obj) throws IOException {
								if (obj instanceof HttpReference) {
									try {
										return createProxy((HttpReference) obj);
									} catch (ClassNotFoundException e) {
										throw new IOException("Could not create proxy: "+e);
									}
								}
								return super.resolveObject(obj);
							}
						};
						try {
							// First boolean is a discriminator - true - success, false - exception
							if (ois.readBoolean()) {
								return ois.readObject();
							} else {
								throw (Exception) ois.readObject();
							}
						} finally {
							ois.close();
						}
					}
					throw new HttpRemoteException("Remoting problem, HTTP error code: "+statusCode+", "+httpMethod.getResponseBodyAsString());
			} finally {
				httpMethod.releaseConnection();
			}							
		} catch (HttpRemoteException e) {
			throw e;					
		} catch (Exception e) {
			throw new HttpRemoteException("Remoting problem: "+e, e);
		}
	}
	
	/**
	 * Creates proxy from reference
	 * @param reference
	 * @return
	 * @throws ClassNotFoundException 
	 */
	protected Object createProxy(final HttpReference reference) throws ClassNotFoundException {
		final HttpRemoteInvocationHandler ih = new HttpRemoteInvocationHandler(
				reference, 
				cookie, 
				connectionManager, 
				classLoader);
		
		class RemoteClassLoader extends ClassLoader {

			public RemoteClassLoader(ClassLoader parent) {
				super(parent);
			}

			// TODO load resources from remote location, cache, define classes, cache.
		}
		
		RemoteClassLoader remoteClassLoader = new RemoteClassLoader(classLoader==null ? getClass().getClassLoader() : classLoader);
		
		Class[] interfaces = new Class[reference.getInterfaces().length]; 
		for (int i=0; i<interfaces.length; ++i) {
			interfaces[i] = remoteClassLoader.loadClass(reference.getInterfaces()[i]);
		}
		
		return Proxy.newProxyInstance(remoteClassLoader, interfaces, ih);				
		
	}
		
}
